﻿local CLootMaster    = LibStub("AceAddon-3.0"):GetAddon("CouncilLootMaster")

local AceGUI = LibStub("AceGUI-3.0")

local L = LibStub("AceLocale-3.0"):GetLocale("CouncilLootMaster")

local LOOTBUTTON_MAXNUM = 10
local LOOTBUTTON_HEIGHT = 32
local LOOTBUTTON_PADDING = 6

-- Column for the scrollingTable
local gearBgColor   = {["r"] = 0.15, ["g"] = 0.15, ["b"] = 0.15, ["a"] = 1.0 }

local sstScrollCols = {
	{ ["name"] = "C.",					["width"] = 20,  ["align"] = "CENTER", ["sortnext"] = 4 },
	{ ["name"] = L["Candidate"],		["width"] = 100, ["align"] = "LEFT" },
	{ ["name"] = L["Rank"],				["width"] = 100, ["align"] = "LEFT" },
	{ ["name"] = L["Response"],			["width"] = 100, ["align"] = "LEFT", 	  	["defaultsort"] = "dsc", ["color"] = {["r"] = 0.25, ["g"] = 1.00, ["b"] = 0.25, ["a"] = 1.0 }, ["sort"]="dsc", ["sortnext"]=2 },

	{ ["name"] = L["Votes"],		    ["width"] = 45,  ["align"] = "RIGHT"},

	{ ["name"] = L["Note"],				["width"] = 30,  ["align"] = "RIGHT"},

	{ ["name"] = "",					["width"] = 5, 	 ["align"] = "LEFT"},

	{ ["name"] = L["iLvl"],				["width"] = 70,  ["align"] = "CENTER", 	 	["bgcolor"] = gearBgColor },
	{ ["name"] = "s1",		          	["width"] = 20,  ["align"] = "CENTER", 		["bgcolor"] = gearBgColor },
	{ ["name"] = "s2",		          	["width"] = 20,  ["align"] = "CENTER",  	["bgcolor"] = gearBgColor },
	{ ["name"] = " ",		          	["width"] = 5,   ["align"] = "LEFT",    	["bgcolor"] = gearBgColor }
}

local enchScrollCols = {
	{ ["name"] = "Enchanters", ["width"] = 100, ["align"] = "LEFT", ["sort"] = "asc" }
}

-- Filters PASS responses from candidate table
local DoFilter = function(self,row)
	if row.response==CLootMaster.RESPONSE.PASS then
		return false
	else
		return true
	end
end

-- Filter for candidate table that passes every candidate
local DontFilter = function(self, ...)
	return true
end

--[[
	Shows UI tooltips
]]
function CLootMasterML:ShowInfoPopup( ... )
	GameTooltip:SetOwner(self.frame, "ANCHOR_NONE")
	for i=1,select("#", ...) do
		if i==1 then
			GameTooltip:AddLine( tostring(select(i, ...)), 1, 1, 1 )
		else
			GameTooltip:AddLine( tostring(select(i, ...)), nil, nil, nil, true )
		end
	end
	GameTooltip:Show()
	GameTooltip:ClearAllPoints()
	if self.frame:IsShown() then
		GameTooltip:SetPoint("TOPLEFT", self.frame , "TOPRIGHT", 0, 0)
	else
		GameTooltip:SetPoint("TOPLEFT", self.frame.titleFrame , "BOTTOMLEFT", 0, 0)
	end
end

--[[
	Hides UI tooltips
]]
function CLootMasterML:HideInfoPopup()
	GameTooltip:Hide()
end

--[[
	Returns the council frame, and creates it if it didnt exist yet
]]
function CLootMasterML:GetFrame()

	if self.frame then
		return self.frame
	end

	local mainframe = CreateFrame("Frame","CLootMasterMLMainFrame",UIParent)
	mainframe:SetPoint("CENTER",UIParent,"CENTER",0,0)
	mainframe:Hide()
	mainframe:SetScale(CLootMaster.db.profile.mainUIScale or 1)
	mainframe:SetMovable(true)
	mainframe:SetFrameStrata("DIALOG")
	self.mainframe = mainframe


	local frame = CreateFrame("Frame","CLootMasterMLFrame",mainframe)
	frame:Show()
	frame:SetPoint("TOPLEFT",mainframe,"TOPLEFT",0,0)
	frame:SetPoint("BOTTOMRIGHT",mainframe,"BOTTOMRIGHT",0,0)
	frame:SetWidth(790)
	frame:SetHeight(355)
	frame:EnableMouse()
	frame:SetMovable(true)
	frame:SetBackdrop({
		bgFile = "Interface\\DialogFrame\\UI-DialogBox-Gold-Background",
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
		tile = true, tileSize = 64, edgeSize = 16,
		insets = { left = 4, right = 4, top = 4, bottom = 4 }
	})
	frame:SetBackdropColor(0.35,0.35,0.35,1)

	local extralootframe = CreateFrame("Frame","CLootMasterMLFrameExtraLoot",frame)
	extralootframe:SetBackdrop({
		bgFile = "Interface\\DialogFrame\\UI-DialogBox-Gold-Background",
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
		tile = true, tileSize = 64, edgeSize = 8,
		insets = { left = 2, right = 1, top = 2, bottom = 2 }
	})
	extralootframe:SetBackdropBorderColor(1,1,1,0.5)
	extralootframe:SetBackdropColor(0.35,0.35,0.35,1)
	extralootframe:Show()
	extralootframe:SetFrameStrata("HIGH")
	extralootframe:SetPoint("TOPRIGHT",frame,"TOPLEFT",4,-10)
	extralootframe:SetWidth(LOOTBUTTON_HEIGHT + 17 )
	frame.extralootframe = extralootframe

	local titleFrame = CreateFrame("Frame", nil, mainframe)
	titleFrame:SetBackdrop({
		bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
		tile = true, tileSize = 64, edgeSize = 8,
		insets = { left = 2, right = 1, top = 2, bottom = 2 }
	})
	titleFrame:SetBackdropColor(0,0,0,1)
	titleFrame:SetHeight(22)
	titleFrame:EnableMouse()
	titleFrame:EnableMouseWheel(true)
	titleFrame:SetPoint("LEFT",mainframe,"TOPLEFT",20,0)
	titleFrame:SetPoint("RIGHT",mainframe,"TOPRIGHT",-20,0)
	titleFrame:SetFrameLevel(3)
	titleFrame:SetToplevel(true)

	titleFrame:SetScript("OnMouseDown", function() mainframe:StartMoving() end)
	titleFrame:SetScript("OnMouseWheel", function(s, delta)
		self:SetUIScale( max(min(mainframe:GetScale(0.8) + delta/15,2.0),0.5) )
	end)
	titleFrame:SetScript("OnEnter", function() self:ShowInfoPopup("CouncilLootmaster", L["Click and drag to move this window."], L["Doubleclick to fold/unfold this window."]) end)
	titleFrame:SetScript("OnLeave", self.HideInfoPopup)
	titleFrame:SetScript("OnMouseUp", function()
		mainframe:StopMovingOrSizing()
		if mainframe.lastClick and GetTime()-mainframe.lastClick<=0.5 then
			if frame:IsShown() then
				frame:Hide()
				titleFrame:ClearAllPoints()
				titleFrame:SetPoint("CENTER",mainframe,"TOP",0,0)
				titleFrame:SetWidth( titleFrame.titletext:GetWidth() + 20 )
				self:HideInfoPopup()
			else
				frame:Show()
				titleFrame:ClearAllPoints()
				titleFrame:SetPoint("LEFT",mainframe,"TOPLEFT",20,0)
				titleFrame:SetPoint("RIGHT",mainframe,"TOPRIGHT",-20,0)
				self:HideInfoPopup()
			end
			mainframe.lastClick = nil
		else
			mainframe.lastClick = GetTime()
		end
	end)

	local titletext = titleFrame:CreateFontString(nil,"OVERLAY","GameFontNormalSmall")
	titletext:SetPoint("CENTER",titleFrame,"CENTER",0,1)
	titletext:SetText( string.format("CouncilLootMaster by Lykeon, based on EPGPLootMaster", self:GetVersionString() ) )
	titleFrame.titletext = titletext
	frame.titleFrame = titleFrame

	local icon = CreateFrame("Button", "COUNCILLM_CURRENTITEMICON", frame, "AutoCastShineTemplate")
	icon:EnableMouse()
	icon:SetNormalTexture("Interface/ICONS/INV_Misc_QuestionMark")
	icon:SetScript("OnEnter", function()
		if not frame.currentLoot then return end
		GameTooltip:SetOwner(frame, "ANCHOR_NONE")
		GameTooltip:SetHyperlink( frame.currentLoot.link )
		GameTooltip:ClearAllPoints()
		GameTooltip:SetPoint("TOPLEFT", frame , "TOPRIGHT", 0, -5)
		GameTooltip:Show()
	end)
	icon:SetScript("OnLeave", function()
		GameTooltip:Hide()
	end)
	icon:SetScript("OnClick", function()
		if not frame.currentLoot then return end
		if ( IsModifiedClick() ) then
			HandleModifiedItemClick(frame.currentLoot.link)
		end
	end)
	icon:SetPoint("TOPLEFT",frame,"TOPLEFT",10,-20)
	icon:SetHeight(48)
	icon:SetWidth(48)
	frame.itemIcon = icon

	local lblItem = frame:CreateFontString(nil,"OVERLAY","GameFontNormalLarge")
	lblItem:SetPoint("TOPLEFT",icon,"TOPRIGHT",10,0)
	lblItem:SetVertexColor( 1, 1, 1 )
	lblItem:SetText( "Itemname" )
	frame.lblItem = lblItem

	local lblInfo = frame:CreateFontString(nil,"OVERLAY","GameFontNormalSmall")
	lblInfo:SetPoint("BOTTOMLEFT",lblItem,"BOTTOMRIGHT",5,0)
	lblInfo:SetVertexColor( 0.7, 0.7, 0.7 )
	lblInfo:SetText( "ItemInfo" )
	frame.lblInfo = lblInfo

	local equipHeaderFrame = CreateFrame("Frame", nil, frame)
	equipHeaderFrame:SetBackdrop({
		bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
		tile = true, tileSize = 64, edgeSize = 8,
		insets = { left = 2, right = 1, top = 2, bottom = 2 }
	})
	equipHeaderFrame:SetBackdropColor(0.2,0.2,0.2,0.6)
	equipHeaderFrame:SetWidth(120)
	equipHeaderFrame:SetHeight(38)

	local titletext = equipHeaderFrame:CreateFontString(nil,"OVERLAY","GameFontNormalSmall")
	titletext:SetVertexColor( 0.9, 0.9, 0.9 )
	titletext:SetPoint("CENTER",equipHeaderFrame,"CENTER",0,0)
	titletext:SetPoint("TOP",equipHeaderFrame,"TOP",0,-5)
	titletext:SetText( string.format("Current equipment:", self:GetVersionString() ) )

	local sstScroll = ScrollingTable:CreateST(sstScrollCols, 12, 20, nil, frame)
	sstScroll.frame:SetPoint("TOPLEFT",frame,"TOPLEFT",10,-95)

	equipHeaderFrame:SetPoint("BOTTOMRIGHT",sstScroll.frame,"TOPRIGHT",-4,-5)

	frame:SetMinResize(frame:GetWidth(),130)
	frame:SetMaxResize(frame:GetWidth(), 60*15+85 )

	frame.sstScroll = sstScroll

	local enchScroll = ScrollingTable:CreateST(enchScrollCols, 10, 20, nil, frame)
	enchScroll.frame:SetPoint("TOPLEFT",sstScroll.frame,"TOPRIGHT",4,0)
	frame.enchScroll = enchScroll

	local lblNoDistribute = frame:CreateFontString(nil,"OVERLAY","GameFontNormal")
	lblNoDistribute:SetVertexColor( 1, 0, 0 )
	lblNoDistribute:SetPoint("TOPLEFT",lblItem,"BOTTOMLEFT",0,-15)
	lblNoDistribute:SetText( L["** MONITOR ONLY **"] )
	frame.lblNoDistribute = lblNoDistribute

	local btnAnnounce = CreateFrame("Button", nil, frame, "UIPanelButtonTemplate")
	btnAnnounce:SetScript("OnEnter", function() self:ShowInfoPopup( L["Announce loot"],
																	L["Click to announce this item to all candidates"],
																	L["This will open the selecton screen on their client."]) end)
	btnAnnounce:SetScript("OnLeave", self.HideInfoPopup)
	btnAnnounce:SetPoint("LEFT",icon,"RIGHT",10,0)
	btnAnnounce:SetHeight(25)
	btnAnnounce:SetText(L["Announce loot"])
	btnAnnounce:SetWidth(btnAnnounce:GetFontString():GetStringWidth() + 20)
	frame.btnAnnounce = btnAnnounce

	local btnDiscard = CreateFrame("Button", nil, frame, "UIPanelButtonTemplate")
	btnDiscard:SetScript("OnEnter", function() self:ShowInfoPopup( L["Discard loot"],
																L["Click to remove this item and all the candidate selections from your list."],
																L["Use this when you don't want to loot this item and close the lootmaster window."]) end)
	btnDiscard:SetScript("OnLeave", self.HideInfoPopup)
	btnDiscard:SetPoint("TOPRIGHT",frame,"TOPRIGHT",-10,-10)
	btnDiscard:SetHeight(25)
	btnDiscard:SetText(L["Discard loot"])

	local btnShowPass = CreateFrame("Button", nil, frame, "UIPanelButtonTemplate")
	btnShowPass:SetScript("OnEnter", function() self:ShowInfoPopup(	"Show passes",
																	"Shows/hides the people that passed on the current item.") end)
	btnShowPass:SetScript("OnLeave", self.HideInfoPopup)
	btnShowPass:SetPoint("TOPRIGHT",btnDiscard,"BOTTOMRIGHT",0,-5)
	btnShowPass:SetHeight(25)
	btnShowPass:SetText("Show Passers")
	local maxwidth = math.max(btnDiscard:GetFontString():GetStringWidth(),btnShowPass:GetFontString():GetStringWidth())
	btnDiscard:SetWidth(maxwidth + 20)
	btnShowPass:SetWidth(maxwidth + 20)

	frame.btnDiscard = btnDiscard
	frame.btnShowPass = btnShowPass

	local lblDisenchant = frame:CreateFontString(nil,"OVERLAY","GameFontNormalSmall")
	lblDisenchant:SetPoint("TOP",enchScroll.frame,"BOTTOM",0,-2)
	lblDisenchant:SetText("Current: no one")
	frame.lblDisenchant = lblDisenchant

	local btnDisenchant = CreateFrame("Button", nil, frame, "UIPanelButtonTemplate")
	btnDisenchant:SetScript("OnEnter", function() self:ShowInfoPopup( "Disenchant Loot",
																"Click to send this item to the enchanter selected above") end)
	btnDisenchant:SetScript("OnLeave", self.HideInfoPopup)
	btnDisenchant:SetPoint("TOP",lblDisenchant,"BOTTOM",0,0)
	btnDisenchant:SetHeight(25)
	btnDisenchant:SetText("Disenchant")
	btnDisenchant:SetWidth(btnDisenchant:GetFontString():GetStringWidth() + 20)
	frame.btnDisenchant = btnDisenchant

	btnAnnounce:SetScript("OnClick", function()
		if not frame.currentLoot then return message(L['no loot selected']) end
		self:AnnounceLoot( frame.currentLoot.extItemID )
		btnAnnounce:Hide()
	end)
	btnDiscard:SetScript("OnClick", function()
		if frame.currentLoot then
			self:RemoveLoot( frame.currentLoot.extItemID )
		end
		self:UpdateUI()
	end)
	btnShowPass:SetScript("OnClick", function()
		if frame.currentLoot then
			if btnShowPass:GetText()=="Show Passers" then
				btnShowPass:SetText("Hide Passers")
				sstScroll:SetFilter(DontFilter)
			else
				btnShowPass:SetText("Show Passers")
				sstScroll:SetFilter(DoFilter)
			end
		end
	end)
	btnDisenchant:SetScript("OnClick", function()
		if frame.currentLoot then
			if self.currentDE then
				self:GiveLootToCandidate(frame.currentLoot.extItemID, self.currentDE, CLootMaster.LOOTTYPE.DISENCHANT )
			else
				self:Print("Loot not sent for DE, no disenchanter selected.")
			end
		end
	end)

	local drop = CreateFrame("Frame", "CLootMasterMLCandidateDropDown", frame, "UIDropDownMenuTemplate")
	drop.addon = self
	drop:SetID(1)
	UIDropDownMenu_Initialize(drop, function(...) CLootMasterML.CandidateDropDownInitialize(CLootMasterML, ...) end, "MENU")
	self.CandidateDropDown = drop

	self.frame = frame

	mainframe:SetHeight( frame:GetHeight() )
	self:UpdateWidth()

	return self.frame
end

--[[
	Function to set the scale of the UI
]]
function CLootMasterML:SetUIScale( scale )
	CLootMaster.db.profile.mainUIScale = scale
	if not self.mainframe then return end
	self.mainframe:SetScale( scale )
end

local extraWidthToggle = true

--[[
	Updates frame width for current contents
]]
function CLootMasterML:UpdateWidth()
	if not self.frame or not self.mainframe or not self.frame.sstScroll then return nil end

	local enchWidth = 0

	if self.frame.enchScroll.frame:IsShown() then
		enchWidth = 5 + self.frame.enchScroll.frame:GetWidth()
	end

	if self.frame.sstScroll.data and #self.frame.sstScroll.data>15 then
		if not extraWidthToggle then
			self.frame:SetWidth( self.frame.sstScroll.frame:GetWidth() + enchWidth + 37 )
			extraWidthToggle = true
		end
	else
		if extraWidthToggle then
			self.frame:SetWidth( self.frame.sstScroll.frame:GetWidth() + enchWidth + 19 )
			extraWidthToggle = false
		end
	end
	self.mainframe:SetWidth( self.frame:GetWidth() )
end

--[[
	If the option is enabled, hide the UI in combat
]]
function CLootMasterML:EnterCombat()
	-- Should we hide when entering combat?
	if not CLootMaster.db.profile.hideMLOnCombat then return end

	self.inCombat = true
	if self:IsShown() then
		self:Hide()
		self.hiddenOnCombat = true
	end
end

--[[
	When leaving combat, show the UI if required
]]
function CLootMasterML:LeaveCombat()
	self.inCombat = nil

	if not self.lootTable then return end

	-- We left combat, see if theres still loot in the cache.
	for id, loot in pairs(self.lootTable) do
		if loot then
			self.hiddenOnCombat = nil
			self:ReloadMLTableForLoot( id )
			break
		end
	end
end

--[[
	Returns true when the UI is shown
]]
function CLootMasterML:IsShown()
	if not self.mainframe then return false end

	if self.inCombat and self.hiddenOnCombat then return true end

	return self.mainframe:IsShown()
end

--[[
	Shows the current council frame
]]
function CLootMasterML:Show()
	if not self.mainframe then return false end

	-- Lootmaster ui is shown... do nothing.
	if CLootMaster.db.profile.hideOnSelection and CLootMaster and CLootMaster.IsShown and CLootMaster.IsShown(CLootMaster) then
		return self.mainframe:Hide()
	end

	-- Are we in combat?
	if self.inCombat then
		-- Show the ui after combat
		self.mainframe:Hide()
		self.hiddenOnCombat = true
		return true
	end

	local ret = self.mainframe:Show()

	if self.frame.currentLoot then
		self:UpdateUI(self.frame.currentLoot.link)
	else
		self:UpdateUI()
	end

	return ret
end

--[[
	Hides the current council frame
]]
function CLootMasterML:Hide()
	if not self.mainframe then return end
	return self.mainframe:Hide()
end

--[[
	Reload the scrollingtable if the current viewing item == link
]]
function CLootMasterML:ReloadMLTableForLoot( itemName )

	-- Lootmaster Loot UI visible? don't update
	if CLootMaster.db.profile.hideOnSelection and CLootMaster and CLootMaster.IsShown and CLootMaster.IsShown(CLootMaster) then return end

	local frame = self:GetFrame()
	if not self.mainframe:IsShown() then
		self.mainframe:Show()
	end

	local lootData = self:GetLoot( itemName )
	if not lootData then
		return self:Print( tostring(itemName) .. ' not found when updating UI')
	end

	self:UpdateUI( lootData.link )

end

local numTotalLootButtons = 0

--[[
	Creates a loot button for selection of loot
]]
function CLootMasterML:CreateLootButton()

	numTotalLootButtons = numTotalLootButtons+1

	local icon = CreateFrame("Button", "COUNCILMLLootButton"..numTotalLootButtons, self.frame, "AutoCastShineTemplate")
	icon:EnableMouse()
	icon:SetNormalTexture("Interface/ICONS/INV_Misc_QuestionMark")
	icon:SetScript("OnEnter", function()
		if not icon.data then return end
		GameTooltip:SetOwner(self.frame, "ANCHOR_NONE")
		GameTooltip:SetHyperlink( icon.data.link )
		GameTooltip:Show()
		GameTooltip:SetPoint("TOPLEFT", self.frame , "TOPRIGHT", 0, -5)
	end)
	icon:SetScript("OnLeave", function()
		GameTooltip:Hide()
	end)
	icon:SetScript("OnClick", function()
		if not icon.data then return end
		if ( IsModifiedClick() ) then
			HandleModifiedItemClick(icon.data.link)
		else
			self:DisplayLoot(icon.data.link)
			self:UpdateUI()
		end
	end)
	icon:SetPoint("RIGHT",self.frame,"LEFT",-5,0)
	icon:SetHeight(LOOTBUTTON_HEIGHT)
	icon:SetWidth(LOOTBUTTON_HEIGHT)
	
	local iconText = icon:CreateFontString(nil, "OVERLAY", "NumberFontNormal")
	iconText:SetPoint("CENTER",icon,"CENTER")
	icon.iconText = iconText
	
	return icon
end

--[[
	Updates the council frame to display the given item
]]
function CLootMasterML:DisplayLoot( item )

	local data = self:GetLoot(item)

	if not data then
		self.frame.currentLoot = nil
		return
	end

	local isCurrentItem = false
	if self.frame.currentLoot and data.link==self.frame.currentLoot.link then
		isCurrentItem = true
	end

	self.frame.currentLoot = data

	if data.quantity>1 then
		self.frame.lblItem:SetText(format('%sx %s',data.quantity, data.link or 'nil'))
	else
		self.frame.lblItem:SetText(data.link or 'nil')
	end

	local color = ITEM_QUALITY_COLORS[data.quality]
	if not color then
		color = {['r']=1,['g']=1,['b']=1}
	end
	self.frame.lblItem:SetVertexColor(color.r, color.g, color.b)

	local binding = ''
	if data.binding=='use' then
		binding = L[', BoU']
	elseif data.binding=='pickup' then
		binding = L[', BoP']
	elseif data.binding=='equip' then
		binding = L[', BoE']
	end

	self.frame.lblInfo:SetText(format("ilevel: %s%s", data.ilevel or -1, binding or ''))
	self.frame.itemIcon:SetNormalTexture(data.texture)

	if not data.announced then
		self.frame.btnAnnounce:Show()
	else
		self.frame.btnAnnounce:Hide()
	end

	if data.mayDistribute then
		self.frame.lblNoDistribute:Hide()
		self.frame.enchScroll:Show()
		self.frame.btnDisenchant:Show()
		self.frame.lblDisenchant:Show()
	else
		self.frame.lblNoDistribute:Show()
		self.frame.lblNoDistribute:SetText( format(L["** MONITORING ** Only %s may distribute this item **"], tostring(data.lootmaster)) )
		self.frame.enchScroll:Hide()
		self.frame.btnDisenchant:Hide()
		self.frame.lblDisenchant:Hide()

		-- clear enchanter table
		self.enchanterTable = {
			['enchanters'] = {},
			['rowdata'] = {}
		}
	end

	if not isCurrentItem then
		self.frame.sstScroll:SetData( data.rowdata )
		if self.frame.btnShowPass:GetText()=="Show Passers" then
			self.frame.sstScroll:SetFilter(DoFilter)
		else
			self.frame.sstScroll:SetFilter(DontFilter)
		end
		-- Restore the default sorting when we're displaying a new item.
		local cols = self.frame.sstScroll.cols
		for i=1, #cols do
		self.frame.sstScroll.cols[i].sort = cols[i].defaultsort
		end
		if self.enchanterTable['enchanters'] then
			self.frame.enchScroll:SetData( self.enchanterTable.rowdata )
		end
	end
	self.frame.sstScroll:SortData()
	self.frame.enchScroll:SortData()
	self.frame.sstScroll:DoFilter()
end

--[[
	Updates the UI given current data
]]
function CLootMasterML:UpdateUI( updateItemLink )

	if not self:IsShown() then return self:Debug('UpdateUI: not shown') end

	local visibleLootButtons = 0
	local totalLoot = 0

	if CLootMaster.db.profile.hideOnSelection and CLootMaster and CLootMaster.IsShown and CLootMaster.IsShown(CLootMaster) then return false end

	-- Are we in combat?
	if self.inCombat then
		-- Show the ui after combat
		self.mainframe:Hide()
		self.hiddenOnCombat = true
		return true
	end

	local breakMe = false

	for item, data in pairs(self.lootTable) do repeat
		if item and data then
			-- If monitoring is disabled just don't display items we're not allowed to distribute.
			if not CLootMaster.db.profile.monitor and not data.mayDistribute then
				-- also just remove the loot
				self:RemoveLoot(item)
				break
			end

			-- If item quantity<1, just remove the loot.
			if data.quantity<1 then
				self:RemoveLoot(item)
				break
			end

			totalLoot = totalLoot + 1

			if visibleLootButtons>=LOOTBUTTON_MAXNUM then breakMe = true; break end
			visibleLootButtons = visibleLootButtons + 1

			if not self.lootButtons then self.lootButtons = {} end

			local lootButton = self.lootButtons[visibleLootButtons]
			if not lootButton then
				lootButton = self:CreateLootButton()
				self.lootButtons[visibleLootButtons] = lootButton
			end

			lootButton.data = data
			lootButton:SetNormalTexture(data.texture)
			lootButton.iconText:SetText(data.ilevel)
			lootButton.iconText:SetWidth(lootButton.iconText:GetStringWidth())

			if not data.voteRequired then
				AutoCastShine_AutoCastStart(lootButton)
			else
				AutoCastShine_AutoCastStop(lootButton)
			end

			lootButton:Show()

			lootButton:SetPoint( "TOP", self.frame, "TOP", 0, -20 - ((LOOTBUTTON_HEIGHT+LOOTBUTTON_PADDING) * (visibleLootButtons-1)) )

			if self.frame.currentLoot and data.link==self.frame.currentLoot.link then
				-- If already displaying, do nothing.
				self:DisplayLoot(item)
				lootButton:LockHighlight()
			elseif not self.frame.currentLoot then
				-- Nothing is onscreen, display the first item
				self:DisplayLoot(item)
				lootButton:LockHighlight()
			else
				lootButton:UnlockHighlight()
			end
		end
	until true
		if breakMe then
			breakMe = false
		end
	end

	if breakMe then
		self:Print("Break doesn't work as expected, contact Author!")
	end

	if self.lootButtons then
		for i = visibleLootButtons+1, LOOTBUTTON_MAXNUM do
			local lootButton = self.lootButtons[i]
			if lootButton then
				AutoCastShine_AutoCastStop(lootButton)
				lootButton.data = nil
				lootButton:Hide()
			end
		end
	end

	if visibleLootButtons>0 then
		self.frame.extralootframe:Show()
		self.frame.extralootframe:SetHeight( (LOOTBUTTON_HEIGHT+LOOTBUTTON_PADDING)*visibleLootButtons + 15 )
	else
		self.frame.extralootframe:Hide()
	end

	if totalLoot==0 and self:IsShown() then
		self.frame.currentLoot = nil
		self:Hide()
	else
		self:UpdateWidth()
	end

end

--[[
	Function to create the right click candidate dropdown list
]]
function CLootMasterML:CandidateDropDownInitialize( frame, level, menuList )

	if not CLootMasterML.CandidateDropDown then return end

	local loot = CLootMasterML.GetLoot( CLootMasterML, CLootMasterML.CandidateDropDown.selectedLink )

	if not loot then
		CLootMasterML:Print(CLootMasterML, L['could not display lootdropdown; loot not in table'])
		return frame:Hide()
	end


	local info = UIDropDownMenu_CreateInfo()

	if UIDROPDOWNMENU_MENU_LEVEL == 1 then

		if CLootMasterML and CLootMasterML.CandidateDropDown and CLootMasterML.CandidateDropDown.selectedCandidate then
			info.notCheckable = 1
			info.isTitle = true
			info.disabled = false
			info.text = CLootMasterML.CandidateDropDown.selectedCandidate
			info.tooltipTitle = nil
			info.tooltipText = nil
			UIDropDownMenu_AddButton(info,UIDROPDOWNMENU_MENU_LEVEL)
			info=UIDropDownMenu_CreateInfo()
			info.notCheckable = 1
		end

		if not (CLootMaster.db.profile.votingDisallowSelf and CLootMaster.IsItMe(CLootMasterML.CandidateDropDown.selectedCandidate)) then
			info.notCheckable = 1
			info.disabled = false
			info.text = L["Vote"]
			info.tooltipTitle = L["Vote"]
			info.tooltipText = L["Click to vote for this candidate"]
			info.func = function() CLootMasterML.CastVote(CLootMasterML, CLootMasterML.CandidateDropDown.selectedLink, CLootMasterML.CandidateDropDown.selectedCandidate) end
			UIDropDownMenu_AddButton(info,UIDROPDOWNMENU_MENU_LEVEL)
		end

		info.notCheckable = 1
		info.disabled = false
		info.text = L['Whisper']
		info.tooltipTitle = L['Whisper']
		info.tooltipText = L['Send a message to the selected candidate.']
		info.func = function() ChatFrame_SendTell(CLootMasterML.CandidateDropDown.selectedCandidate) end
		UIDropDownMenu_AddButton(info,UIDROPDOWNMENU_MENU_LEVEL)

		if loot.mayDistribute then
			info.isTitle = true
			info.text = ''
			info.disabled = false
			info.tooltipTitle = nil
			info.tooltipText = nil
			UIDropDownMenu_AddButton(info,UIDROPDOWNMENU_MENU_LEVEL)
			info=UIDropDownMenu_CreateInfo()
			info.notCheckable = 1

			if not loot.manual then
				info.isTitle = false
				info.text = 'Give loot'
				info.tooltipTitle = info.text
				info.tooltipText = 'Attempts to send the loot to the candidate for winning the vote'
				info.func = function() CLootMasterML.GiveLootToCandidate(CLootMasterML, CLootMasterML.CandidateDropDown.selectedLink, CLootMasterML.CandidateDropDown.selectedCandidate, CLootMaster.LOOTTYPE.VOTE) end
				UIDropDownMenu_AddButton(info,UIDROPDOWNMENU_MENU_LEVEL)
			else
				info.isTitle = false
				info.disabled = false
				info.text = L['- Cannot distribute loot -']
				info.tooltipTitle = info.text
				info.tooltipText = L["You have added this loot manually to the list, you will need to handle the loot manually and discard the loot from the list when you're done distributing it."]
				info.func = function() end
				UIDropDownMenu_AddButton(info,UIDROPDOWNMENU_MENU_LEVEL)
			end

			info.isTitle = true
			info.disabled = false
			info.text = ''
			info.tooltipTitle = nil
			info.tooltipText = nil
			UIDropDownMenu_AddButton(info,UIDROPDOWNMENU_MENU_LEVEL)
			info=UIDropDownMenu_CreateInfo()
			info.notCheckable = 1

			info.isTitle = false
			info.disabled = false
			info.text = L['(Re)announce loot to candidate']
			info.tooltipTitle = L['(Re)announce loot to candidate']
			info.tooltipText = L['Reopens the loot selection popup at the candidate, this offers the candidate to vote for the loot after a crash or disconnect.']
			info.func = function() CLootMasterML.AskCandidateIfNeeded(CLootMasterML, CLootMasterML.CandidateDropDown.selectedLink, CLootMasterML.CandidateDropDown.selectedCandidate) end
			UIDropDownMenu_AddButton(info,UIDROPDOWNMENU_MENU_LEVEL)

			info.isTitle = false
			info.disabled = false
			info.hasArrow = 1
			info.text = L['Set response manually']
			info.tooltipTitle = L['Set response manually']
			info.tooltipText = L['Allows you to manually set the response for a given candidate']
			info.func = function() end
			info.value = 'RESPONSE_OVERRIDE'
			UIDropDownMenu_AddButton(info,UIDROPDOWNMENU_MENU_LEVEL)
		end

	elseif UIDROPDOWNMENU_MENU_LEVEL==2 then

		if UIDROPDOWNMENU_MENU_VALUE == 'RESPONSE_OVERRIDE' then
			local respList = {"MAIN","OFF"}
			for i=1, 2 do
				info.isTitle = false
				info.disabled = false
				info.text = CLootMaster.RESPONSE[respList[i].."_TEXT"]
				info.tooltipTitle = CLootMaster.RESPONSE[respList[i].."_TEXT"]
				info.tooltipText = L['Manually sets the response of this candidate to ']..CLootMaster.RESPONSE[respList[i].."_TEXT"]..L['. Please note that the candidate will receive a notice about this in whisper.']
				info.func = function()
					CLootMasterML.SetManualResponse(CLootMasterML, CLootMasterML.CandidateDropDown.selectedLink, CLootMasterML.CandidateDropDown.selectedCandidate, CLootMaster.RESPONSE[respList[i]] )
					CloseDropDownMenus()
				end
				local ddBtn = UIDropDownMenu_AddButton(info,UIDROPDOWNMENU_MENU_LEVEL)
			end
			info.isTitle = false
			info.disabled = false
			info.text = L['Pass']
			info.tooltipTitle = L['Pass']
			info.tooltipText = L['Manually sets the response of this candidate to pass. Please note that the candidate will receive a notice about this in whisper.']
			info.func = function()
				CLootMasterML.SetManualResponse(CLootMasterML, CLootMasterML.CandidateDropDown.selectedLink, CLootMasterML.CandidateDropDown.selectedCandidate, CLootMaster.RESPONSE.PASS )
				CloseDropDownMenus()
			end
			UIDropDownMenu_AddButton(info,UIDROPDOWNMENU_MENU_LEVEL)
		end
	end
end

--[[
	Function called when right clicked on a candidate in the scrollingtable
]]
function CLootMasterML:OnCandidateRowRightClick( candidate, link, row )
	self.CandidateDropDown.selectedCandidate = candidate
	self.CandidateDropDown.selectedLink = link
	self.CandidateDropDown.selectedRow = row

	ToggleDropDownMenu(1, nil, self.CandidateDropDown, "cursor", 0, 0)
end

--[[
	Sets the disenchanter the Disenchant button sends items to
]]
function CLootMasterML:SetDisenchanter(enchanter)
	if not enchanter then
		self.currentDE = nil
		self.frame.lblDisenchant:SetText("Current: no one")
	else
		self.currentDE = enchanter
		self.frame.lblDisenchant:SetText("Current: "..CLootMaster.Ambiguate(enchanter,"none"))
	end
end

--[[
	Draw function for empty cells in scrollingtable
]]
function CLootMasterML:EmptyCellOwnerDraw(cell)
	if not cell then return end
	cell.text:SetText('')
end

--[[
	Draw function for note button in scrollingtable
]]
function CLootMasterML:SetNoteCellOwnerDraw(cell, itemData)

	cell.text:SetText('')

	if not itemData or itemData=='' then
		cell.CouncilLMModTexture = false
		return cell:SetNormalTexture(nil)
	end
	
	-- Automatic note text scanner TODO: expand functionality
	if string.lower(itemData)=='bis' then
		cell.CouncilLMModTexture = false
		cell.text:SetText('BiS')
		cell.text:SetTextColor(0,0.8,0)
		return cell:SetNormalTexture(nil)
	end

	cell:SetNormalTexture("Interface\\Buttons\\UI-GuildButton-PublicNote-Up")

	-- Center the texture if not done already.
	if not cell.CouncilLMModTexture then
		local t = cell:GetNormalTexture()
		t:ClearAllPoints()
		t:SetPoint("CENTER",t:GetParent(),"CENTER")
		cell.CouncilLMModTexture = true
	end

end

--[[
	Draw function for current gear in scrollingtable
]]
function CLootMasterML:SetGearCellOwnerDraw(cell, itemData)

	if not itemData or itemData=='' then
		cell.text:SetText('')
		cell:SetNormalTexture(nil)
		return
	end

	local link, ilevel, itemTexture = strsplit("^", itemData)

	cell.text:SetText('')
	cell:SetNormalTexture(itemTexture)
end

--[[
	Draw function for candidate name (due to having to ambiguate)
]]
function CLootMasterML:SetCandidateNameCellOwnerDraw(cell, itemData)
	if not itemData or itemData=='' then
        cell.text:SetText('');
        cell:SetNormalTexture(nil)
        return;
    end

	cell.text:SetText(CLootMaster.Ambiguate(itemData,"none"))
end

--[[
	Draw function for enchanter name (due to having to ambiguate)
]]
function CLootMasterML:SetEnchanterNameCellOwnerDraw(cell, itemData)
	if not itemData or itemData=='' then
        cell.text:SetText('');
        cell:SetNormalTexture(nil)
        return;
    end

	cell.text:SetText(CLootMaster.Ambiguate(itemData,"none"))
end

--[[
	Draw function for class icons in scrollingtable
]]
function CLootMasterML:SetClassIconCellOwnerDraw(cell, itemData)

	cell.text:SetText('')

	if not itemData or itemData=='' or not CLASS_ICON_TCOORDS[itemData] then
		cell:SetNormalTexture(nil)
		return
	end

	local coords = CLASS_ICON_TCOORDS[itemData]
	cell:SetNormalTexture("Interface\\Glues\\CharacterCreate\\UI-CharacterCreate-Classes")
	cell:GetNormalTexture():SetTexCoord(coords[1],coords[2],coords[3],coords[4])
end

--[[
	Shows tooltip for finding candidate current gear if unknown
]]
function CLootMasterML:ShowGearInspectPopup( candidate, item )
	local foundGear = self:GetCandidateData(item, candidate, "foundGear") or false
	if not foundGear then
		self:ShowInfoPopup( candidate, L['Click to retrieve current equipment.'] )
	end
end

--[[
	Shows note popup for when note button is pressed in scrollingtable
]]
function CLootMasterML:ShowNoteCellPopup( candidate, item )
	local itemData = self:GetCandidateData(item, candidate, 'note')
	if not itemData or itemData=='' then return end -- TODO for extended note functionality filter non-note stuff here

	self:ShowInfoPopup(L['Note added by '] .. candidate .. ':', itemData or '')
end

--[[
	Shows tooltip for candidate current gear
]]
function CLootMasterML:ShowGearCellPopup( candidate, item, dataName )

	local itemData = self:GetCandidateData(item, candidate, dataName)
	if not itemData or itemData=='' then return end

	local link, ilevel, itemTexture = strsplit("^", itemData)

	GameTooltip:SetOwner(self.frame, "ANCHOR_NONE")
	GameTooltip:SetHyperlink( link )
	GameTooltip:Show()
	GameTooltip:SetPoint("TOPLEFT", self.frame , "TOPRIGHT", 0, -5)

end

--[[
	Shows candidate tooltip (as if you moused over them)
]]
function CLootMasterML:ShowCandidateCellPopup( candidate, item )
	if not candidate then return nil end
	GameTooltip:SetOwner(self.frame, "ANCHOR_NONE")
	GameTooltip:SetUnit(CLootMaster.Ambiguate(candidate,'none'))
	GameTooltip:Show()
	GameTooltip:SetPoint("TOPLEFT", self.frame , "TOPRIGHT", 0, -5)
end

--[[
	Shows tooltip with voting information
]]
function CLootMasterML:ShowVotesCellPopup( candidate, item )
	local loot = self:GetLoot(item)
	if not loot then return end

	if not loot.voteRequired then
		-- build a list of people who voted for the selected candidate
		local voters = nil
		if type(loot.voters) == 'table' then
			-- ok, we've voted for this item, see if theres another item on the list that needs a vote!
			for listVoter, listCandidate in pairs(loot.voters) do
				if candidate==listCandidate then
					if voters==nil then voters={} end
					tinsert(voters, listVoter)
				end
			end
		end

		if voters==nil or #(voters)<1 then
			-- Noone voted for this candidate, poor guy/gal
			self:ShowInfoPopup(	L['Votes'],
								format(L['Noone has voted for %s'], candidate))
		else
			-- Someone voted for this candidate, lucky bastard!
			self:ShowInfoPopup(	L['Votes'],
								format(L['The following players have voted for %s:'], candidate))
			for i, voter in ipairs(voters) do
				GameTooltip:AddLine( voter, 1, 1, 1 )
			end
			GameTooltip:Show()
		end
	end
end

--[[
	Sets candidates average item level text in scrollingtable
]]
function CLootMasterML:SetGearCelliLVL( cell, self, candidate, item )

	local avgItemLvl = self:GetCandidateData(item, candidate, "avgItemLvl") or 0

	cell.text:SetText( avgItemLvl )
end

--[[
	Returns an RGB color corresponding to a given candidate's response for a given item
]]
function CLootMasterML:GetCandidateCellColor(candidate, item, dataName, defaultColor)

	local r, g, b = self:GetCandidateResponseColor( candidate, item, nil )
	return {["r"] = r or 1, ["g"] = g or 0, ["b"] = b or 1, ["a"] = 1.0 }
end

--[[
	Returns an RGB color corresponding to the given candidate's class
]]
function CLootMasterML:GetCandidateClassCellColor( candidate, item, dataName, defaultColor )
	local color = RAID_CLASS_COLORS[self:GetCandidateData(item, candidate, "unitclass")]
	if not color then
		-- if class not found display epic color.
		color = {["r"] = 0.63921568627451, ["g"] = 0.2078431372549, ["b"] = 0.93333333333333, ["a"] = 1.0 }
	else
		color.a = 1.0
	end
	return color
end

--[[
	Returns an RGB color corresponding to the given enchanter's class
]]
function CLootMasterML:GetEnchanterClassCellColor( enchanter )
	local color = RAID_CLASS_COLORS[self:GetEnchanterData(enchanter, "unitclass")]
	if not color then
		-- if class not found display epic color.
		color = {["r"] = 0.63921568627451, ["g"] = 0.2078431372549, ["b"] = 0.93333333333333, ["a"] = 1.0 }
	else
		color.a = 1.0
	end
	return color
end

--[[
	Returns the color for a given candidate's response to a given item
]]
function CLootMasterML:GetCandidateResponseColor( candidate, item, response )
	if not response then
		response = tonumber(self:GetCandidateData(item, candidate, "response"))
	end

	if not response or not CLootMaster.RESPONSE[response] then
		return 1,0,1
	end

	return unpack( CLootMaster.RESPONSE[response].COLOR or {1,0,1} )
end

--[[
	Sets text corresponding to candidate's response in scrollingtable
]]
function CLootMasterML:SetCandidateResponseCellUserDraw( cell, self, candidate, item )

	local response = tonumber(self:GetCandidateData(item, candidate, "response"))

	local text = nil

	if CLootMaster.RESPONSE[response] then
		text = CLootMaster.RESPONSE[response].TEXT

	else
		text = 'resp: ' .. self:GetCandidateData(item, candidate, "response")
	end

	-- Add looted status message when candidate has looted the item.
	if self:GetCandidateData(item, candidate, "looted") then
		return cell.text:SetText( (text or '')  .. '; ' .. L['Looted'] )
	end

	return cell.text:SetText( text or '' )
end

--[[
	Sets vote area for a candidate in the scrollingtable,
	shows voting button when your vote is needed, otherwise
	the number of votes for that candidate
]]
function CLootMasterML:SetCandidateVotesCellUserDraw( cell, self, candidate, item )
	local loot = self:GetLoot(item)

	local v = tonumber(self:GetCandidateData(loot.extItemID, candidate, "votes")) or 0

	if loot.voteRequired and not (CLootMaster.db.profile.votingDisallowSelf and CLootMaster.IsItMe(candidate)) then
		-- show button
		cell.text:SetText('')

		-- create button if it doesn't exist
		if not cell.votebutton then
			local btn = CreateFrame("Button", nil, cell, "UIPanelButtonTemplate")
			btn:SetPoint("TOPLEFT",cell,"TOPLEFT",0,0)
			btn:SetPoint("BOTTOMRIGHT",cell,"BOTTOMRIGHT",0,0)
			btn:SetText(L["Vote"])
			cell.votebutton = btn
			btn:SetScript("OnClick", function()
				if not cell.candidate or not cell.item then
					print('no data!')
					return
				end
				self:CastVote(cell.item, cell.candidate)
			end)
			btn:SetScript("OnEnter", function()
				self:ShowInfoPopup(	L['Votes'],
									format(L['Please click this button to vote for %s for the selected item'], cell.candidate))
			end)
			btn:SetScript("OnLeave", self.HideInfoPopup)
		end

		cell.item = item
		cell.candidate = candidate
		cell.votebutton:Show()
	else
		-- hide button (if available)
		if cell.votebutton then cell.votebutton:Hide() end

		if v>0 and not loot.voteRequired then
			cell.text:SetText(v)
		else
			cell.text:SetText('')
		end
	end
end

--[[
	Hides gear popup
]]
function CLootMasterML:HideGearCellPopup( candidate, item, dataName )
	GameTooltip:Hide()
end

--[[
	Re-sorts the table when inspecting a candidate
]]
function CLootMasterML:OnGearInspectClick( candidate, item )
	if self:InspectCandidate( item, candidate ) then
		self.frame.sstScroll:SortData()
		self.frame.sstScroll:DoFilter()
	else
		self:Print( format(L['%s is offline, out of range or not grouped, unable to inspect.'], candidate or 'Unknown') )
	end
end

--[[
	Makes WoW handle modified clicks on gear icons in scrolling table
	be handled like modified clicks on an item link
]]
function CLootMasterML:OnGearCellClick( candidate, item, dataName )
	if ( IsModifiedClick() ) then
		local link, _, _, _, _ = strsplit("^", self:GetCandidateData(item, candidate, dataName) or '')
		HandleModifiedItemClick(link)
	end
end